from abc import ABCMeta
from copy import copy
from typing import Callable

from .config import RegistryConfig


class _DictMixin:
    """Dict-like methods for a registry-based class."""

    __registry__: dict
    __registry_config__: RegistryConfig

    def __getitem__(self, key: str):
        return self.__registry_config__.getitem(self.__registry__, key)

    def __iter__(self):
        for val in self.__registry__:
            yield val

    def __len__(self):
        return len(self.__registry__)

    def __contains__(self, val: str):
        val = self.__registry_config__.format(val)
        return val in self.__registry__

    def keys(self):
        return self.__registry__.keys()

    def values(self):
        return self.__registry__.values()

    def items(self):
        for item in self.__registry__.items():
            yield item

    def get(self, key: str, default=None):
        try:
            return self[key]
        except KeyError:
            pass
        if isinstance(default, str):
            return self[default]
        else:
            return default

    def clear(self):
        self.__registry__.clear()


class RegistryMeta(ABCMeta, _DictMixin):
    __registry__: dict
    __registry_config__: RegistryConfig

    def __new__(mcls, name, bases, namespace, **config):
        cls = super().__new__(mcls, name, bases, namespace)

        # Each subclass gets its own registry.
        cls.__registry__ = {}

        try:
            Registry
        except NameError:
            # Should only happen the very first time that
            # Registry is being defined.
            cls.__registry_config__ = RegistryConfig(**config)
            return cls

        # Copy the nearest parent config, then update it with new params
        for parent_cls in cls.mro()[1:]:
            try:
                cls.__registry_config__ = copy(parent_cls.__registry_config__)  # type: ignore
                cls.__registry_config__.update(config)
                break
            except AttributeError:
                pass

        # Register direct subclasses of Register to Register
        if cls in Registry.__subclasses__() and name != "RegistryDecorator":
            Registry.__registry_config__.register(Registry.__registry__, cls)

        # otherwise, register it in own registry and all parent registries.
        for parent_cls in cls.mro():
            if parent_cls == Registry:
                continue

            if (
                not cls.__registry_config__.recursive
                and cls not in parent_cls.__subclasses__()
            ):
                continue

            try:
                config = parent_cls.__registry_config__  # type: ignore
            except AttributeError:
                # Not a Registry object
                continue

            if not config.register_self and parent_cls == cls:
                continue

            config.register(parent_cls.__registry__, cls)  # type: ignore

        return cls

    def __repr__(self):
        return f"<{self.__name__}: {list(self.keys())}>"


class Registry(metaclass=RegistryMeta):
    __call__: Callable  # For decorating
    __getitem__: Callable
    __len__: Callable
    __contains__: Callable
    keys: Callable
    values: Callable
    items: Callable
    get: Callable
    clear: Callable

    def __new__(cls, *args, **kwargs):
        if cls is Registry:
            # A Registry is being explicitly created for decorating
            return super().__new__(RegistryDecorator, *args, **kwargs)
        else:
            # Registry is being subclassed
            return super().__new__(cls)


class RegistryDecorator(Registry, _DictMixin):
    def __init__(self, /, **config):
        """Create a Registry for decorating."""
        # overwrite the registry datas so its independent
        # of the Registry object.
        self.__registry__ = {}
        self.__registry_config__ = RegistryConfig(**config)

    def __call__(self, obj):
        """For decorator."""
        self.__registry_config__.register(self.__registry__, obj)
        return obj

    def __repr__(self):
        return f"<Registry: {list(self.__registry__.keys())}>"



# Don't manually change, let poetry-dynamic-versioning-plugin handle it.
__version__ = "0.2.0"

from abc import abstractmethod

from .config import InvalidNameError, RegistryConfig
from .registry import Registry, RegistryDecorator, RegistryMeta



import re
from dataclasses import dataclass
from typing import Callable


class InvalidNameError(Exception):
    """"""


_to_snake_case_pattern1 = re.compile("(.)([A-Z][a-z]+)")
_to_snake_case_pattern2 = re.compile("__([A-Z])")
_to_snake_case_pattern3 = re.compile("([a-z0-9])([A-Z])")


def to_snake_case(name: str) -> str:
    """Convert PascalCase to snake_case.

    Based on:
        https://stackoverflow.com/a/1176023

    Parameters
    ----------
    name : str

    Returns
    -------
    str
    """
    name = _to_snake_case_pattern1.sub(r"\1_\2", name)
    name = _to_snake_case_pattern2.sub(r"_\1", name)
    name = _to_snake_case_pattern3.sub(r"\1_\2", name)
    return name.lower()


@dataclass
class RegistryConfig:
    case_sensitive: bool = False
    suffix: str = ""
    strip_suffix: bool = True

    # Classes only; Register to its own registry
    register_self: bool = False

    # If ``True``, register new entries in all parent registries.
    # Otherwise, only register in parent.
    recursive: bool = True

    snake_case: bool = False

    def update(self, new: dict):
        for key, value in new.items():
            if hasattr(self, key):
                setattr(self, key, value)

    def getitem(self, registry: dict, key: str):
        if not self.case_sensitive:
            key = key.lower()

        return registry[key]

    def register(self, registry: dict, func: Callable):
        name = func.__name__

        if not name.endswith(self.suffix):
            raise InvalidNameError(f'"{func}" name must end with "{self.suffix}"')

        if self.strip_suffix and self.suffix:
            name = name[: -len(self.suffix)]

        name = self.format(name)

        registry[name] = func

    def format(self, name):
        if self.snake_case:
            name = to_snake_case(name)

        if not self.case_sensitive:
            name = name.lower()

        return name



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['autoregistry']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'autoregistry',
    'version': '0.2.0',
    'description': 'Automatic registry design-pattern for mapping names to functionality.',
    'long_description': '.. image:: https://raw.githubusercontent.com/BrianPugh/autoregistry/main/assets/logo_400w.png\n\n|Python compat| |GHA tests| |Codecov report| |readthedocs|\n\n.. inclusion-marker-do-not-remove\n\nAutoRegistry\n============\n\nInvoking functions and class-constructors from a string is a common design pattern\nthat ``autoregistry`` aims to solve.\nFor example, a user might specify a backend of type ``"sqlite"`` in a yaml configuration\nfile, for which our program needs to construct the ``SQLite`` subclass of our ``Database`` class.\nClassically, you would need to manually create a lookup mapping the string ``"sqlite"`` to\nthe ``SQLite`` constructor.\nWith ``autoregistry``, the lookup is autmatically created for you.\n\n\n``autoregistry`` has a single  powerful class ``Registry`` that can do the following:\n\n* Be subclassed to automatically register subclasses by their name.\n\n* Be directly invoked ``my_registery = Registry()`` to create a decorator\n  for registering callables like functions.\n\n.. inclusion-marker-remove\n\nInstallation\n============\nAutoRegistry requires Python ``>=3.8``.\n\n.. code-block:: bash\n\n   python -m pip install autoregistry\n\n\nExamples\n========\n\nClass Inheritence\n^^^^^^^^^^^^^^^^^\n\n``Registry`` adds a dictionary-like interface to class constructors\nfor looking up subclasses.\n\n.. code-block:: python\n\n   from dataclasses import dataclass\n   from autoregistry import Registry, abstractmethod\n\n\n   @dataclass\n   class Pokemon(Registry):\n       level: int\n       hp: int\n\n       @abstractmethod\n       def attack(self, target):\n           """Attack another Pokemon."""\n\n\n   class Charmander(Pokemon):\n       def attack(self, target):\n           return 1\n\n\n   class Pikachu(Pokemon):\n       def attack(self, target):\n           return 2\n\n\n   class SurfingPikachu(Pikachu):\n       def attack(self, target):\n           return 3\n\n\n   print(f"{len(Pokemon)} Pokemon registered:")\n   print(f"    {list(Pokemon)}")\n   charmander = Pokemon["cHaRmAnDer"](\n       level=7, hp=31\n   )  # By default, lookup is case-insensitive\n   print(f"Created Pokemon: {charmander}")\n\nThis code block produces the following output:\n\n.. code-block::\n\n   3 Pokemon registered:\n       [\'charmander\', \'pikachu\', \'surfingpikachu\']\n   Created Pokemon: Charmander(level=7, hp=31)\n\n\nFunction Registry\n^^^^^^^^^^^^^^^^^\n\nDirectly instantiating a ``Registry`` object allows you to\nregister functions by decorating them.\n\n.. code-block:: python\n\n   from autoregistry import Registry\n\n   pokeballs = Registry()\n\n\n   @pokeballs\n   def masterball(target):\n       return 1.0\n\n\n   @pokeballs\n   def pokeball(target):\n       return 0.1\n\n\n   for ball in ["pokeball", "masterball"]:\n       success_rate = pokeballs[ball](None)\n       print(f"Ash used {ball} and had {success_rate=}")\n\nThis code block produces the following output:\n\n.. code-block:: text\n\n   Ash used pokeball and had success_rate=0.1\n   Ash used masterball and had success_rate=1.0\n\n\n.. |GHA tests| image:: https://github.com/BrianPugh/autoregistry/workflows/tests/badge.svg\n   :target: https://github.com/BrianPugh/autoregistry/actions?query=workflow%3Atests\n   :alt: GHA Status\n.. |Codecov report| image:: https://codecov.io/github/BrianPugh/autoregistry/coverage.svg?branch=main\n   :target: https://codecov.io/github/BrianPugh/autoregistry?branch=main\n   :alt: Coverage\n.. |readthedocs| image:: https://readthedocs.org/projects/autoregistry/badge/?version=latest\n        :target: https://autoregistry.readthedocs.io/en/latest/?badge=latest\n        :alt: Documentation Status\n.. |Python compat| image:: https://img.shields.io/badge/>=python-3.8-blue.svg\n',
    'author': 'Brian Pugh',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
